@@ -10,7 +10,7 @@ import traceback
 
 from azure.storage.queue import QueueClient
 from azure.storage.blob import BlobClient
-from azure.cosmosdb.table import TableService
+from azure.cosmosdb.table import TableService, TableBatch
 from typing import List
 from multiprocessing.pool import ThreadPool
 
@@ -18,6 +18,7 @@ enough_workers_message = "There are already enough active workers"
 workers_long_running = "Workers are active for too long. Going to process as well."
 state_table = "workerState"
 failure_table = "executionFailures"
+segment_failure_table = "segmentExecutionFailures"
 wake_up_queue_name = "wake-up"
 
 input_container = "input"
@@ -176,9 +177,9 @@ class Worker:
         blob_name, params, offset, retry_count, batch_count, next_chunk = self._read_metadata(request_entity)
         try:
             if next_chunk:
-                result, next_chunk = self._process_next_chunk(next_chunk, blob_name, func, offset, params)
+                result, failures, next_chunk = self._process_next_chunk(next_chunk, blob_name, func, offset, params)
             else:
-                result, next_chunk = self._process_blob_by_segments(blob_name, func, params)
+                result, failures, next_chunk = self._process_blob_by_segments(blob_name, func, params)
         except Exception as e:
             if retry_count == 3:
                 self._record_failure(request_id, e)
@@ -187,11 +188,14 @@ class Worker:
                 self.table_client.update_entity(self.metadata_table_name, request_entity)
                 self._enqueue(request_id)
             return
+
+        if failures:
+            self._record_segment_failures(request_id, blob_name, failures)
             
         serialized_result = "\n".join(json.dumps(r) for r in result)
         if next_chunk:
             out_blob = self._get_output_blob("%s/%s_%d" % (self.worker_name, blob_name, batch_count))
-            out_blob.upload_blob(serialized_result, timeout=1200)
+            out_blob.upload_blob(serialized_result, timeout=1200, overwrite=True)
                 
             request_entity['Offset'] = offset + len(result)
             request_entity['Batch'] = batch_count + 1
@@ -201,7 +205,7 @@ class Worker:
         else:
             out_blob = self._get_output_blob("%s/%s" % (self.worker_name, blob_name))
             if offset == 0:
-                out_blob.upload_blob(serialized_result, timeout=1200)
+                out_blob.upload_blob(serialized_result, timeout=1200, overwrite=True)
             else:     
                 content = []
                 for batch in range(batch_count):
@@ -209,7 +213,7 @@ class Worker:
                     blob_content = in_blob.download_blob().readall().decode()
                     content.append(blob_content)
                 content.append(serialized_result)
-                out_blob.upload_blob("\n".join(content), timeout=1200)
+                out_blob.upload_blob("\n".join(content), timeout=1200, overwrite=True)
             request_entity['State'] = "Completed"
             self.table_client.update_entity(self.metadata_table_name, request_entity)
 
@@ -235,18 +239,18 @@ class Worker:
         return blob_name, params, offset, retry_count, batch_count, next_chunk
 
     def _process_blob_by_segments(self, blob_name, func, params):
-        result, batch = [], []
+        batch = []
         in_blob = self._get_input_blob(blob_name)
         blobs_iterator = self._read_blob_by_segments(in_blob)
         for source_segment in blobs_iterator:
             batch.append(source_segment)
             if len(batch) == self.segment_limit:
-                result.extend(func(batch, params))
+                result, failures = func(batch, params)
                 batch = []
                 break
 
         if len(batch) > 0: # If segments in blob are less than segment_limit
-            result.extend(func(batch, params))
+            result, failures = func(batch, params)
 
         next_chunk = []
         for source_segment in blobs_iterator:
@@ -254,13 +258,16 @@ class Worker:
             if len(next_chunk) == self.segment_limit:
                 break
 
-        return result, next_chunk
+        return result, failures, next_chunk
 
     def _process_next_chunk(self, next_chunk, blob_name, func, offset, params):
         chunk_read_task = self.pool.apply_async(self._get_next_batch, args=(offset + len(next_chunk), blob_name)) # Start reading new chunk
-        result = func(next_chunk, params) # Process current chunk
+        result, failures = func(next_chunk, params) # Process current chunk
+        for i in range(len(failures)):
+            s, e, reason = failures[i]
+            failures[i] = (s + offset, e + offset, reason)
         next_chunk = chunk_read_task.get()
-        return result, next_chunk
+        return result, failures, next_chunk
 
     def _get_next_batch(self, offset, blob_name):
         in_blob = self._get_input_blob(blob_name)
@@ -282,6 +289,20 @@ class Worker:
         function_failure_entity = {"PartitionKey": self.worker_name, "RowKey": request_id, "FailureMessage": str(e), "StackTrace": track}
         TableService(connection_string=self.ltb_connection_string).insert_entity(failure_table, function_failure_entity)
 
+    def _record_segment_failures(self, request_id, blob_name, failures):
+        entities = []
+        batch, count = TableBatch(), 0
+        table = TableService(connection_string=self.ltb_connection_string)
+        for start, end, reason in failures:
+            entity = {"PartitionKey": "%s_%s" % (self.worker_name, request_id), "RowKey": "%d_%d" % (start, end), "FailureMessage": reason, "Start": start, "End": end}
+            batch.insert_or_replace_entity( entity)
+            count += 1
+            if count == 99:
+                table.commit_batch(segment_failure_table, batch)
+                batch = TableBatch()
+        if count > 0:
+            table.commit_batch(segment_failure_table, batch)
+
     def _record_request_failure(self, request_entity, e: Exception):
         trace = traceback.format_exc()
         error_message = str(e)
